/*
 * TismTool © Tis Veugen, 2010-2011
 * Simulation, Verification
 *
 * TsmThread.h
 * version 5
 */
#ifndef TSM_THREAD_H
#define TSM_THREAD_H

#include <boost/thread/mutex.hpp>    
#include <boost/thread/locks.hpp>
#include <boost/thread/thread.hpp>
#include "TsmThreadType.h"
#include "TsmExtQueue.h"
#include "TsmIntQueue.h"
#include "TsmMutex.h"
#include "TsmRzVousSem.h"
#include "TsmTriggerSem.h"
#include "TsmEnvelope.h"
#include "TsmRvState.h"
#include "TsmSupervisor.h"
using namespace std;

namespace TsmRTE
{
  class TsmThread
  {
  public:
    TsmThreadType   mType;
    int             mSequence;
    bool            mbHalt;
    TsmExtQueue*    mExternalQueue;
    TsmIntQueue*    mInternalQueue;

  protected:
    boost::thread*    mThread;
    TsmMutex*         mExclusiveMutex;
    TsmRzVousSem      mRendezVousSem;
    TsmTriggerSem     mTriggerSem;
    boost::mutex      mLockMutex;
    boost::thread::id mRendezVousThreadBid;
    string            mName;
    TsmTrigger*       mPendingTrigger;

  public:
    TsmThread( TsmThreadType Type = THREAD );
    virtual ~TsmThread();

    boost::thread* getThread()
    {
      return mThread;
    }

    TsmExtQueue* getExtQueue()
    {
      return mExternalQueue;
    }

    TsmIntQueue* getIntQueue()
    {
      return mInternalQueue;
    }

    int getMaxExtQueueSize()
    {
      return mExternalQueue->getMaxQueueSize();
    }

    int getMaxIntQueueSize()
    {
      return mInternalQueue->getMaxQueueSize();
    }

    bool hasMessage()
    {
      boost::unique_lock<boost::mutex> lock( mLockMutex );
      return ( mExternalQueue->size() + mInternalQueue->size() > 0 );
    }

    string MaxQueueSizes()
    {
      stringstream ss;
      ss.width(3);
      ss << this->getMaxExtQueueSize() << "  ";
      ss.width(3);
      ss << this->getMaxIntQueueSize() << "  ";
      ss << this->getName();
      return ss.str();
    }

    virtual void Prepare( string sName );

    virtual void Prepare()
    {
    }

    virtual void Run()
    {
      TriggerNotify();
    }

    virtual void Halt();

    virtual void Interrupt()
    {
      mThread->interrupt();
    }

    virtual void Cease()
    {
    }

    void ExclusiveLock()
    {
      mExclusiveMutex->WaitOne();
    }

    void ExclusiveUnlock()
    {
      mExclusiveMutex->ReleaseMutex();
    }

    void ExclusiveReset()
    {
      delete mExclusiveMutex;
      mExclusiveMutex = new TsmMutex();
    }

    void RendezVousWait()
    {
      mRendezVousThreadBid = boost::this_thread::get_id();
      mRendezVousSem.WaitOne();
    }

    void RendezVousNotify()
    {
      Tsm::Seq2( "R,", mExternalQueue->getSeqId() );
      TsmSupervisor::SetState( mRendezVousThreadBid );
      mRendezVousSem.Release();
    }

    void TriggerWait()
    {
      mTriggerSem.WaitOne();
    }

    void TriggerNotify()
    {
      mTriggerSem.Release();
    }

    void ASyncCall( TsmEnvelope& oEnvelope );
    void SyncCall( TsmEnvelope& oEnvelope );
    TsmRvState Handle( TsmEnvelope& oEnvelope );

    virtual void NativeCall( TsmEnvelope& oEnvelope )
    {
      // To be overridden!
    }

    virtual void CbHandle( TsmEnvelope& oEnvelope )
    {
      // To be overridden!
    }

    void TsmReset1()
    {
      mExternalQueue->Empty();
      mInternalQueue->Empty();
    }

    void TsmReset2()
    {
      mRendezVousSem.Reset();
      ExclusiveReset();
      if ( mPendingTrigger != NULL )
      {
        delete mPendingTrigger;
        mPendingTrigger = NULL;
      }
    }

    string Snap()
    {
      stringstream sb;

      if ( mExternalQueue->size() > 0 )
      {
        sb << " Q:";
      }
      sb << mExternalQueue->Snap();

      if ( mInternalQueue->size() > 0 )
      {
        sb << " S:";
      }
      sb << mInternalQueue->Snap();

      return sb.str();
    }

    string getName()
    {
      return mName;
    }

    void setName( string s )
    {
      mName = s;
    }
  };

  void Worker( TsmThread* oTsmThread );

}
#endif
